#ifdef RCSID
static char RCSid[] =
"$Header$";
#endif

/* Copyright (c) 1999 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  w32tdbt3.cpp - HTML TADS Debugger - TADS 3 engine configuration
Function
  
Notes
  
Modified
  11/24/99 MJRoberts  - Creation
*/

#include <Windows.h>

#include <string.h>
#include <ctype.h>

#include "os.h"

#include "vmvsn.h"

#include "t3main.h"
#include "t3std.h"
#include "w32main.h"
#include "w32tdb.h"
#include "w32prj.h"
#include "tadsdock.h"
#include "htmldcfg.h"
#include "w32dbgpr.h"
#include "w32abt.h"
#include "tadsdlg.h"
#include "tadsapp.h"
#include "htmldcfg.h"

/* from w32t3.cpp */
extern char w32t3_credits_txt[];

/* ------------------------------------------------------------------------ */
/*
 *   Debugger cover function for beginning program execution.  This
 *   routine re-builds the argument vector from the arguments stored in
 *   the configuration settings.  
 */
static int t3_dbg_main(int argc, char **argv, struct appctxdef *appctx,
                       char *config_file)
{
    CHtmlSys_dbgmain *dbgmain = CHtmlSys_dbgmain::get_main_win();
    char **my_argv;
    int my_argc;
    int i;
    char *src;
    char *dst;
    char logfile[OSFNMAX];
    char cmdfile[OSFNMAX];
    char args[1024];
    CHtmlDebugConfig *config;
    char quote;
    int ret;

    /* 
     *   if there's no main debugger window, we can't synthesize any
     *   arguments 
     */
    if (dbgmain == 0)
        return t3main(argc, argv, appctx, config_file);

    /* get the local configuration object */
    config = dbgmain->get_local_config();

    /*
     *   Ignore the command line passed in from the caller, and use our
     *   own command line instead.  First, retrieve the command line from
     *   the configuration settings.  
     */
    if (config->getval("prog_args", "args", 0, args, sizeof(args)))
        args[0] = '\0';

    /* skip any leading spaces in the argument string */
    for (src = args ; isspace(*src) ; ++src) ;

    /* 
     *   count the first argument, if it is present - we count the start
     *   of each new argument 
     */
    my_argc = 0;
    if (*src != '\0')
        ++my_argc;

    /* 
     *   Parse the arguments into individual strings, null-terminating
     *   each argument, but leaving them in the original buffer.  Note
     *   that the buffer can only shrink, as we remove quotes, but it can
     *   never grow; this allows us to process the buffer in-place,
     *   without making a separate copy.  
     */
    for (quote = '\0', dst = args ; *src != '\0' ; )
    {
        /* if process according to whether we're in a quoted string or not */
        if (quote == '\0')
        {
            /* 
             *   we're not in a quoted string - if this is a space, it
             *   starts the next argument; if it's a quote, note that
             *   we're in a quoted string, and remove the quote; otherwise
             *   just copy it out 
             */
            switch(*src)
            {
            case '"':
                /* it's a quote - note it, but don't copy it to the output */
                quote = *src;
                ++src;
                break;

            case ' ':
                /* 
                 *   it's the end of the current argument - null-terminate
                 *   the current argument string 
                 */
                *dst++ = '\0';

                /* skip the space we just scanned */
                ++src;

                /* skip to the next non-space character */
                while (isspace(*src))
                    ++src;

                /* 
                 *   if another argument follows, count the start of the
                 *   new argument 
                 */
                if (*src != '\0')
                    ++my_argc;
                break;

            default:
                /* simply copy anything else to the output */
                *dst++ = *src++;
                break;
            }
        }
        else
        {
            /* 
             *   if it's our quote, we're done with the quoted section -
             *   unless it's a stuttered quote, in which case we will
             *   simply copy a single quote to the output 
             */
            if (*src == quote)
            {
                /* check for stuttering */
                if (*(src + 1) == quote)
                {
                    /* it's stuttered - make a single copy of the quote */
                    *dst++ = quote;

                    /* skip both copies of the quote in the input */
                    src += 2;
                }
                else
                {
                    /* 
                     *   it's the end of our quoted section - skip it in
                     *   the input, but don't copy it to the output 
                     */
                    ++src;

                    /* note that we're no longer in a quoted section */
                    quote = '\0';
                }
            }
            else
            {
                /* 
                 *   it's just the next character of the quoted section -
                 *   copy it unchanged 
                 */
                *dst++ = *src++;
            }
        }
    }

    /* null-terminate the last argument string */
    *dst = '\0';

    /* get the log file, if one is specified */
    if (config->getval("prog_args", "logfile", 0, logfile, sizeof(logfile)))
        logfile[0] = '\0';

    /* get the command input file, if one is specified in the configuration */
    if (config->getval("prog_args", "cmdfile", 0, cmdfile, sizeof(cmdfile)))
        cmdfile[0] = '\0';

    /* add space for the executable name argument (argv[0]) */
    ++my_argc;

    /* add space for the image file name argument */
    ++my_argc;

    /* 
     *   add space for the log file, if there is one - we need two
     *   arguments for this (one for the "-l", and one for the filename) 
     */
    if (logfile[0] != '\0')
        my_argc += 2;

    /* add another two for the command file, if there is one ("-c file") */
    if (cmdfile[0] != '\0')
        my_argc += 2;

    /* allocate space for the argument vector */
    my_argv = (char **)t3malloc(my_argc * sizeof(*my_argv));

    /* copy the original argv[0] */
    my_argv[0] = argv[0];

    /* start at slot 1 */
    i = 1;

    /* add the log file argument, if present */
    if (logfile[0] != '\0')
    {
        my_argv[i++] = "-l";
        my_argv[i++] = logfile;
    }

    /* add the command file argument, if present */
    if (cmdfile[0] != '\0')
    {
        my_argv[i++] = "-i";
        my_argv[i++] = cmdfile;
    }

    /* add the program name, which we can learn from the debugger */
    my_argv[i++] = (char *)dbgmain->get_gamefile_name();

    /* add the user arguments */
    for (src = args ; i < my_argc ; ++i, src += strlen(src) + 1)
    {
        /* set this argument */
        my_argv[i] = src;
    }

    /* run the program */
    ret = t3main(my_argc, my_argv, appctx, config_file);

    /* free the argument vector */
    t3free(my_argv);

    /* return the result */
    return ret;
}


/* ------------------------------------------------------------------------ */
/*
 *   Define startup configuration variables for the TADS debugger 
 */
int (*w32_tadsmain)(int, char **, struct appctxdef *, char *) = t3_dbg_main;
char *w32_beforeopts = "";
char *w32_configfile = 0;
int w32_allow_debugwin = FALSE;
int w32_always_pause_on_exit = FALSE;
char *w32_setup_reg_val_name = "Debugger Setup Done";
char *w32_usage_app_name = "htmltdb3";
char *w32_titlebar_name = "HTML T3";
int w32_in_debugger = TRUE;
char *w32_opendlg_filter = "T3 Applications\0*.t3x\0All Files\0*.*\0\0";
char *w32_debug_opendlg_filter =
    "T3 Applications, Configurations, and Sources\0"
    "*.t3x;*.tdc;*.t\0"
    "All Files\0*.*\0\0";
char *w32_config_opendlg_filter =
    "T3 Applications and Configurations\0*.t3x;*.tdc\0"
    "All Files\0*.*\0\0";
char *w32_game_ext = "t3x";
const char *w32_version_string = T3VM_VSN_STRING;
const char w32_pref_key_name[] =
    "Software\\TADS\\HTML TADS\\3.0\\Settings";
const char w32_tdb_global_prefs_file[] = "htmltdb3.tdc";

/* ------------------------------------------------------------------------ */
/*
 *   Debugger version factoring routines for TADS 3
 */

/*
 *   Close windows in preparation for loading a new configuration 
 */
void CHtmlSys_dbgmain::vsn_load_config_close_windows()
{
    /* close the project window */
    if (projwin_ != 0)
        SendMessage(GetParent(projwin_->get_handle()), DOCKM_DESTROY, 0, 0);
}

/*
 *   Open windows for a newly-loaded configuration 
 */
void CHtmlSys_dbgmain::vsn_load_config_open_windows(const RECT *deskrc)
{
    RECT rc;
    int vis;
    
    /* create the project window */
    projwin_ = new CHtmlDbgProjWin(this);
    SetRect(&rc, deskrc->left + 2, deskrc->bottom - 2 - 150, 300, 150);
    projwin_->load_win_config(local_config_, "project win", &rc, &vis);
    load_dockwin_config_win(local_config_, "project win", projwin_,
                            "Project", vis, TADS_DOCK_ALIGN_LEFT,
                            170, 270, &rc, TRUE);
}

/*
 *   Save version-specific configuration information 
 */
void CHtmlSys_dbgmain::vsn_save_config()
{
    /* save the project window's layout information */
    if (projwin_ != 0)
    {
        /* save the project window's layout */
        projwin_->save_win_config(local_config_, "project win");
        projwin_->save_win_config(global_config_, "project win");

        /* save the project file information */
        projwin_->save_config(local_config_);
    }
}

/*
 *   check command status 
 */
TadsCmdStat_t CHtmlSys_dbgmain::vsn_check_command(int command_id)
{
    switch(command_id)
    {
    case ID_PROJ_ADDFILE:
    case ID_PROJ_ADDEXTRES:
    case ID_PROJ_ADDRESDIR:
        /* 
         *   always route these to the project window, even if it's not
         *   the active window 
         */
        return (projwin_ != 0
                ? projwin_->check_command(command_id)
                : TADSCMD_UNKNOWN);

    default:
        /* we don't handle other cases */
        return TADSCMD_UNKNOWN;
    }
}

/* 
 *   execute command 
 */
int CHtmlSys_dbgmain::vsn_do_command(int nfy, int cmd, HWND ctl)
{
    switch(cmd)
    {
    case ID_PROJ_ADDFILE:
    case ID_PROJ_ADDEXTRES:
    case ID_PROJ_ADDRESDIR:
        /* 
         *   always route these to the project window, even if it's not
         *   the active window 
         */
        return (projwin_ != 0 ? projwin_->do_command(nfy, cmd, ctl) : FALSE);
        
    default:
        /* we don't handle other cases */
        return FALSE;
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Create a temporary directory.  Returns zero on success, non-zero on
 *   failure. 
 */
static int create_temp_dir(char *dirbuf, HWND owner_hwnd)
{
    char systmp[OSFNMAX];

    /* 
     *   seed the random number generator - we'll try random names until
     *   we find one that doesn't collide with another process's names 
     */
    srand((unsigned int)GetTickCount());

    /* 
     *   get the system temporary directory - we'll create our temporary
     *   directory as a subdirectory 
     */
    GetTempPath(sizeof(systmp), systmp);

    /* keep going until we find a unique name */
    for (;;)
    {
        UINT id;
        
        /* generate a temporary directory name */
        id = (UINT)rand();
        GetTempFileName(systmp, "tads", id, dirbuf);

        /* try creating the directory */
        if (CreateDirectory(dirbuf, 0))
        {
            /* success */
            return 0;
        }

        /* check the error - if it's not "already exists", give up */
        if (GetLastError() != ERROR_ALREADY_EXISTS)
        {
            /* tell the user what's wrong */
            MessageBox(owner_hwnd,
                       "An error occurred creating a temporary "
                       "directory.  Please free up some space on "
                       "your hard disk and try again.", "TADS Workbench",
                       MB_OK | MB_ICONEXCLAMATION);

            /* tell the caller we failed */
            return 1;
        }
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Run the compiler.  We build a list of commands, then we start a
 *   thread to execute the list.  We build the list so that the background
 *   build thread doesn't need to allocate memory, access the
 *   configuration, or otherwise look at any data that aren't thread-safe.
 */
void CHtmlSys_dbgmain::run_compiler(int command_id)
{
    CStringBufCmd cmdline(2048);
    char extres_base_filename[OSFNMAX];
    char exe[OSFNMAX];
    char cmd_dir[OSFNMAX];
    size_t cmd_dir_len;
    char gamefile[OSFNMAX];
    char optfile[OSFNMAX];
    char buf[1024];
    int i;
    int cnt;
    int intval;
    DWORD thread_id;
    int extres_cnt;
    int cur_extres;
    int force_full = FALSE;
    char tmpdir[OSFNMAX];

    /* 
     *   make one last check to make sure we don't already have a build in
     *   progress - we can only handle one at a time 
     */
    if (build_running_)
        return;

    /* presume we won't need a temporary directory */
    tmpdir[0] = '\0';

    /* clear the build command list */
    clear_build_cmds();

    /*
     *   Build the name of the game file.  If we're building a debug game,
     *   use the current game name.  If we're building the release game,
     *   use the release game name.  If we're building an executable,
     *   build to a temporary file. 
     */
    switch(command_id)
    {
    case ID_BUILD_COMPDBG:
    case ID_BUILD_COMPDBG_FULL:
    case ID_BUILD_COMP_AND_RUN:
        /* build to the loaded game file */
        strcpy(gamefile, get_gamefile_name());

        /* we're not going to build external resource files */
        extres_base_filename[0] = '\0';

        /* if this is a full build, add the 'force' flag */
        if (command_id == ID_BUILD_COMPDBG_FULL)
            force_full = TRUE;
        break;

    case ID_BUILD_COMPRLS:
        /* preparing a release - force a full build */
        force_full = TRUE;
        
        /* build to the release game file */
        local_config_->getval("build", "rlsgam", 0,
                              gamefile, sizeof(gamefile));

        /* create a temporary directory for our intermediate files */
        if (create_temp_dir(tmpdir, handle_))
            return;

        /* base the external resource filenames on the release filename */
        strcpy(extres_base_filename, gamefile);
        break;

    case ID_BUILD_COMPEXE:
    case ID_BUILD_COMPINST:
        /* preparing a release - force a full build */
        force_full = TRUE;

        /* create a temporary directory for our intermediate files */
        if (create_temp_dir(tmpdir, handle_))
            return;

        /* 
         *   we're compiling to an executable (this is also the case if
         *   we're building an installer), so we don't really care about
         *   the .gam file; build the game to a temp file 
         */
        GetTempPath(sizeof(buf), buf);
        GetTempFileName(buf, "tads", 0, gamefile);

        /* check to see if we're building an installer */
        if (command_id == ID_BUILD_COMPINST)
        {
            /* 
             *   get the name of the executable - we'll base the external
             *   resource files on this name 
             */
            local_config_->getval("build", "exe", 0, buf, sizeof(buf));

            /* 
             *   put the external resource files into the temp dir, using
             *   the executable's root filename 
             */
            os_build_full_path(extres_base_filename,
                               sizeof(extres_base_filename),
                               tmpdir, os_get_root_name(buf));
        }
        else
        {
            /* 
             *   base the name of the external resource files on the
             *   executable name 
             */
            local_config_->getval("build", "exe", 0, extres_base_filename,
                                  sizeof(extres_base_filename));
        }
        break;
    }

    /* -------------------------------------------------------------------- */
    /*
     *   Step one: run the compiler 
     */

    /* generate the name of the compiler executable */
    GetModuleFileName(0, exe, sizeof(exe));
    strcpy(os_get_root_name(exe), "t3make.exe");

    /* start the command with the executable name */
    cmdline.set("");
    cmdline.append_qu(exe, FALSE);

    /* write intermediate files to our temp dir if we have one */
    if (tmpdir[0] != '\0')
    {
        /* add the symbol file directory option */
        cmdline.append_opt("-Fy", tmpdir);

        /* add the object file directory option */
        cmdline.append_opt("-Fo", tmpdir);
    }

    /* add the include list */
    cnt = local_config_->get_strlist_cnt("build", "includes");
    for (i = 0 ; i < cnt ; ++i)
    {
        /* get this include path */
        local_config_->getval("build", "includes", i, buf, sizeof(buf));

        /* add the -i option and the path */
        cmdline.append_opt("-I", buf);
    }

    /* 
     *   add the output image file name option - enclose the filename in
     *   quotes in case it contains spaces 
     */
    cmdline.append_opt("-o", gamefile);

    /* if we need to force a full build, add the -r option */
    if (force_full)
        cmdline.append_inc(" -r", 512);

    /* add the #defines, enclosing each string in quotes */
    cnt = local_config_->get_strlist_cnt("build", "defines");
    for (i = 0 ; i < cnt ; ++i)
    {
        /* get this include path */
        local_config_->getval("build", "defines", i, buf, sizeof(buf));

        /* add the -D option */
        cmdline.append_opt("-D", buf);
    }

    /* 
     *   add the undefs; these don't need quotes, since they must be
     *   simple identifiers 
     */
    cnt = local_config_->get_strlist_cnt("build", "undefs");
    for (i = 0 ; i < cnt ; ++i)
    {
        /* get this include path */
        local_config_->getval("build", "undefs", i, buf, sizeof(buf));

        /* add the -U option and the symbol to undefine */
        cmdline.append_opt("-U", buf);
    }

#if 0 //$$$$
    /* add the -case- option if required */
    if (!local_config_->getval("build", "ignore case", &intval) && intval)
        cmdline.append_inc(" -case-", 512);

    /* add the -C+ option if required */
    if (!local_config_->getval("build", "c ops", &intval) && intval)
        cmdline.append_inc(" -C+", 512);

    /* add the -ctab option if required */
    if (!local_config_->getval("build", "use charmap", &intval) && intval
        && !local_config_->getval("build", "charmap", 0, buf, sizeof(buf)))
    {
        /* add the option */
        cmdline.append_opt("-ctab", buf);
    }
#endif

    /* add the extra options - just add them exactly as they're entered */
    cnt = local_config_->get_strlist_cnt("build", "extra options");
    for (i = 0 ; i < cnt ; ++i)
    {
        /* get this include path */
        local_config_->getval("build", "extra options", i, buf, sizeof(buf));

        /* add a space, then this line of options */
        cmdline.append_inc(" ", 512);
        cmdline.append_inc(buf, 512);
    }

    /* if we're compiling for debugging, add the debug option */
    if (command_id == ID_BUILD_COMPDBG
        || command_id == ID_BUILD_COMPDBG_FULL
        || command_id == ID_BUILD_COMP_AND_RUN)
        cmdline.append_inc(" -d", 512);

    /* add the list of source files */
    cnt = local_config_->get_strlist_cnt("build", "source_files");
    for (i = 0 ; i < cnt ; ++i)
    {
        /* get this source file */
        local_config_->getval("build", "source_files", i, buf, sizeof(buf));

        /* add it to the command line in quotes */
        cmdline.append_qu(buf, TRUE);
    }

    /* run the compiler in the source file's directory */
    os_get_path_name(cmd_dir, sizeof(cmd_dir), local_config_filename_.get());

    /* add the command to the list */
    add_build_cmd("t3make", exe, cmdline.get(), cmd_dir, FALSE);


    /* -------------------------------------------------------------------- */
    /*
     *   Step two: run the resource bundler.  Bundle resources only for
     *   release/exe builds, not for debugging builds.  
     */
    if (command_id == ID_BUILD_COMPDBG
        || command_id == ID_BUILD_COMPDBG_FULL
        || command_id == ID_BUILD_COMP_AND_RUN)
    {
        /* this is a debug build, so we're done */
        goto done;
    }

    /* build the name of the bundler */
    strcpy(os_get_root_name(exe), "t3res.exe");

    /* start the command with the executable name */
    cmdline.set("");
    cmdline.append_qu(exe, FALSE);

    /* 
     *   run the resource bundler in the project file directory,
     *   regardless of where the game we're actually building will end up 
     */
    GetFullPathName(game_filename_.get(), sizeof(buf), buf, 0);
    os_get_path_name(cmd_dir, sizeof(cmd_dir), buf);

    /* note the length of the command directory, for prefix comparisons */
    cmd_dir_len = strlen(cmd_dir);

    /* add the image file name (preceded by a space, and in quotes) */
    cmdline.append_qu(gamefile, TRUE);

    /* set up the options to add multimedia resources */
    cmdline.append_inc(" -add ", 512);

    /* get the count of resource files to go into the image file */
    cnt = local_config_->get_strlist_cnt("build", "image_resources");

    /* add the image file resources */
    for (i = 0 ; i < cnt ; ++i)
    {
        /* get this source file */
        local_config_->getval("build", "image_resources",
                              i, buf, sizeof(buf));

        /* add it to the command line in quotes */
        cmdline.append_qu(buf, TRUE);
    }

    /* add the build command for the image resources */
    add_build_cmd("t3res", exe, cmdline.get(), cmd_dir, FALSE);

    /* get the number of external resource files */
    if (local_config_->getval("build", "ext resfile cnt", &extres_cnt))
        extres_cnt = 0;

    /* build the external resource files */
    for (cur_extres = 0 ; cur_extres < extres_cnt ; ++cur_extres)
    {
        char config_name[50];

        /* 
         *   check to see if we should include this file, and get its name
         *   if so - if this returns false, it means the file shouldn't be
         *   included, so just skip it in this case 
         */
        if (!build_get_extres_file(buf, cur_extres, command_id,
                                   extres_base_filename))
            continue;

        /* start the command line */
        cmdline.set("");
        cmdline.append_qu(exe, FALSE);

        /* add the resource filename with a -create option */
        cmdline.append_inc(" -create", 512);
        cmdline.append_qu(buf, TRUE);

        /* add the action specifiers */
        cmdline.append_inc(" -add", 512);

        /* generate the name of this external resource file */
        CHtmlDbgProjWin::gen_extres_names(0, config_name, cur_extres);

        /* get the count of resource files to go into the image file */
        cnt = local_config_->get_strlist_cnt("build", config_name);

        /* add the image file resources */
        for (i = 0 ; i < cnt ; ++i)
        {
            /* get this source file */
            local_config_->getval("build", config_name, i, buf, sizeof(buf));
            
            /* add it to the command line in quotes */
            cmdline.append_qu(buf, TRUE);
        }

        /* add the build command */
        add_build_cmd("t3res", exe, cmdline.get(), cmd_dir, FALSE);
    }

    /* -------------------------------------------------------------------- */
    /*
     *   If we're building an executable, run the executable bundler 
     */
    if (command_id == ID_BUILD_COMPINST)
    {
        /* building the installer - skip the executable build step */
        goto build_installer;
    }
    else if (command_id != ID_BUILD_COMPEXE)
    {
        /* we're not building an executable, so we're done */
        goto done;
    }

    /* build the name of the executable builder */
    strcpy(os_get_root_name(exe), "maketrx32.exe");

    /* start the command with the executable name */
    cmdline.set("");
    cmdline.append_qu(exe, FALSE);

    /* specify the source executable (in quotes) */
    GetModuleFileName(0, buf, sizeof(buf));
    strcpy(os_get_root_name(buf), "htmlt3.exe");
    cmdline.append_qu(buf, TRUE);

    /* add the game name */
    cmdline.append_qu(gamefile, TRUE);

    /* add the executable filename in quotes */
    local_config_->getval("build", "exe", 0, buf, sizeof(buf));
    cmdline.append_qu(buf, TRUE);

    /* run the command */
    add_build_cmd("maketrx32", exe, cmdline.get(), cmd_dir, FALSE);

    /* add a command to delete the temporary image file */
    add_build_cmd("del", "del", gamefile, "", TRUE);


    /* -------------------------------------------------------------------- */
    /*
     *   If we're building the installer, run the install maker
     */
build_installer:
    if (command_id != ID_BUILD_COMPINST)
    {
        /* not building the installer - we're done */
        goto done;
    }

    /* build the name of the installer maker */
    strcpy(os_get_root_name(exe), "mksetup.exe");

    /* start the command with the executable name */
    cmdline.set("");
    cmdline.append_qu(exe, FALSE);

    /* 
     *   add the -detached option - this causes the installer creator to
     *   run its own subprocesses as detached, so that we get the results
     *   piped back to us through the standard handles that we create 
     */
    cmdline.append_inc(" -detached", 512);

    /* create a temporary options file name */
    GetTempPath(sizeof(buf), buf);
    GetTempFileName(buf, "tads", 0, optfile);
    
    /* add the options file - this is a temp file */
    cmdline.append_qu(optfile, TRUE);

    /* add the name of the resulting installation program */
    local_config_->getval("build", "installer prog", 0, buf, sizeof(buf));
    cmdline.append_qu(buf, TRUE);

    /* add the command */
    add_build_cmd("mksetup", exe, cmdline.get(), cmd_dir, FALSE);

    /* add a build command to delete the temporary options file */
    add_build_cmd("del", "del", optfile, "", TRUE);

    /* add a build command to delete the temporary game file */
    add_build_cmd("del", "del", gamefile, "", TRUE);

    /* create the options file */
    {
        FILE *fp;
        int cnt;
        int i;
        int found_name;
        int found_savext;
        int found_gam;
        int found_exe;

        /* we haven't found our special values yet */
        found_name = FALSE;
        found_savext = FALSE;
        found_gam = FALSE;
        found_exe = FALSE;

        /* open the file */
        fp = fopen(optfile, "w");

        /* always include the "T3" option */
        fprintf(fp, "T3=1\n");

        /* write the options lines */
        cnt = local_config_->get_strlist_cnt("build", "installer options");
        for (i = 0 ; i < cnt ; ++i)
        {
            /* get this string */
            local_config_->getval("build", "installer options",
                            i, buf, sizeof(buf));

            /* if this is one of the required ones, note its presence */
            if (strlen(buf) > 5 && memicmp(buf, "name=", 5) == 0)
                found_name = TRUE;
            else if (strlen(buf) > 7 && memicmp(buf, "savext=", 7) == 0)
                found_savext = TRUE;
            else if (strlen(buf) > 4 && memicmp(buf, "gam=", 4) == 0)
                found_gam = TRUE;
            else if (strlen(buf) > 4 && memicmp(buf, "exe=", 4) == 0)
                found_exe = TRUE;

            /* write it out */
            fprintf(fp, "%s\n", buf);
        }

        /* write out the "GAM" line if we didn't find an explicit one */
        if (!found_gam)
            fprintf(fp, "GAM=%s\n", gamefile);

        /* write out the "EXE" line if we didn't find an explicit one */
        if (!found_exe)
        {
            /* use the normal build .EXE filename here, sans path */
            local_config_->getval("build", "exe", 0, buf, sizeof(buf));
            fprintf(fp, "EXE=%s\n", os_get_root_name(buf));
        }

        /* add the external resources we're including */
        for (cur_extres = 0 ; cur_extres < extres_cnt ; ++cur_extres)
        {
            /* 
             *   check to see if we should include this file, and get its
             *   name if so - if this returns false, it means the file
             *   shouldn't be included, so just skip it in this case 
             */
            if (!build_get_extres_file(buf, cur_extres, command_id,
                                       extres_base_filename))
                continue;

            /* add this file to the installer options */
            fprintf(fp, "FILE=%s\n", buf);
        }

        /* we've finished writing the options file */
        fclose(fp);

        /* if we didn't find the NAME or SAVEXT settings, it's an error */
        if (!found_name || !found_savext)
        {
            /* show the message */
            MessageBox(handle_, "Before you can build an installer, you "
                       "must fill in the Name and Save Extension options "
                       "in the installer configuration.  Please fill in "
                       "these values, then use Compile Installer again "
                       "when you're done.", "TADS Workbench",
                       MB_OK | MB_ICONEXCLAMATION);

            /* take them to the dialog */
            run_dbg_build_dlg(handle_, this, DLG_BUILD_INSTALL, 0);

            /* 
             *   delete the temporary files (for the options file and for
             *   the game), since we won't actually be carrying out the
             *   build commands, which would ordinarily clean up the temp
             *   files 
             */
            DeleteFile(optfile);
            DeleteFile(gamefile);

            /* delete our temporary directory */
            del_temp_dir(tmpdir);

            /* return without starting a build */
            return;
        }
    }

done:
    /* -------------------------------------------------------------------- */
    /*
     *   We've built the list of commands, so we can now start the build
     *   thread to carry out the command list 
     */

    /* if we have a temporary directory, add a command to delete it */
    if (tmpdir[0] != '\0')
        add_build_cmd("rmdir-s", "rmdir-s", tmpdir, "", TRUE);
    
    /* note that we have a compile running */
    build_running_ = TRUE;

    /* if we're meant to run after the build, so note */
    run_after_build_ = (command_id == ID_BUILD_COMP_AND_RUN);

    /* start the compiler thread */
    build_thread_ = CreateThread(0, 0,
                                 (LPTHREAD_START_ROUTINE)&run_compiler_entry,
                                 this, 0, &thread_id);

    /* if we couldn't start the thread, show an error */
    if (build_thread_ == 0)
    {
        /* note the problem */
        oshtml_dbg_printf("Build failure: unable to start build thread\n");

        /* we don't have a build running after all */
        build_running_ = FALSE;
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Check to determine if it's okay to compile 
 */
int CHtmlSys_dbgmain::vsn_ok_to_compile()
{
    /* make sure the project window has saved its configuration */
    if (projwin_ != 0)
        projwin_->save_config(local_config_);
    
    /* ensure we have at least one source file listed */
    if (local_config_->get_strlist_cnt("build", "source_files") == 0)
    {
        /* 
         *   make sure the project window is visible before showing the
         *   error message, since the message refers to the project 
         */
        SendMessage(get_handle(), WM_COMMAND, ID_VIEW_PROJECT, 0);

        /* tell them what's wrong */
        MessageBox(handle_, "This project has no source files.  Before "
                   "you can compile, you must add at least one "
                   "file to the project's \"Source Files\" group.",
                   "TADS Workbench", MB_OK | MB_ICONEXCLAMATION);

        /* 
         *   make sure the project window is the active window now that
         *   we've removed the error message dialog 
         */
        SendMessage(get_handle(), WM_COMMAND, ID_VIEW_PROJECT, 0);
        
        /* 
         *   don't proceed with compilation right now - wait for the user
         *   to retry explicitly 
         */
        return FALSE;
    }

    /* we can proceed with the compilation */
    return TRUE;
}

/* ------------------------------------------------------------------------ */
/*
 *   Get the name of an external resource file, and check to see if it
 *   should be included in the build 
 */
int CHtmlSys_dbgmain::build_get_extres_file(char *buf, int cur_extres,
                                            int build_command,
                                            const char *extres_base_name)  
{
    char ext[32];

    /* 
     *   if we're building an installer, check to see if this file is
     *   included in the installer, and skip it if not 
     */
    if (build_command == ID_BUILD_COMPINST)
    {
        char in_inst[10];
        
        /* get the in-install status - make it yes by default */
        if (local_config_->getval("build", "ext_resfile_in_install",
                                  cur_extres, in_inst, sizeof(in_inst)))
            in_inst[0] = 'Y';
        
        /* if this one isn't in the install, tell the caller to skip it */
        if (in_inst[0] != 'Y')
            return FALSE;
    }
    
    /* build the name of this external resource file */
    sprintf(ext, "3r%d", cur_extres);
    strcpy(buf, extres_base_name);
    os_remext(buf);
    os_addext(buf, ext);

    /* include the file */
    return TRUE;
}

/* ------------------------------------------------------------------------ */
/*
 *   Create and load a new game 
 */
void CHtmlSys_dbgmain::create_new_game(const textchar_t *srcfile,
                                       const textchar_t *gamefile,
                                       int create_source, int advanced,
                                       const textchar_t *custom_template)
{
    textchar_t new_config[OSFNMAX];
    char msg[512];
    char fbuf[OSFNMAX];
    char rel_path[OSFNMAX];
    char proj_path[OSFNMAX];

    /* first, save the current configuration */
    if (local_config_filename_.get() != 0)
    {
        /* save the current configuration into the configuration object */
        save_config();

        /* write the configuration object to its file */
        local_config_->save_file(local_config_filename_.get());
    }

    /* build the name of the new configuration file */
    strcpy(new_config, gamefile);
    os_remext(new_config);
    os_addext(new_config, "tdc");

    /* note the path name to the project file */
    os_get_path_name(proj_path, sizeof(proj_path), new_config);

    /* 
     *   before we do anything we might regret, make sure we can save the
     *   configuration file 
     */
    if (local_config_->save_file(new_config))
    {
        /* tell them what went wrong */
        sprintf(msg, "An error occurred creating the new game's "
                "configuration file (%s).  Please make sure the folder "
                "is correct and the disk isn't full.", new_config);
        MessageBox(handle_, msg, "TADS Workbench",
                   MB_ICONEXCLAMATION | MB_OK);

        /* stop now with no change to the loaded configuration */
        return;
    }

    /* if desired, create the source file from a starter file */
    if (create_source)
    {
        /* 
         *   use the custom template if provided; otherwise, use one of
         *   the standard templates 
         */
        if (custom_template != 0)
        {
            /* use the specified custom template */
            strcpy(fbuf, custom_template);
        }
        else
        {
            /* 
             *   get the path to the appropriate starter file - it's in
             *   the the TADS program directory, and is called "startA3.t"
             *   for the advanced starter file, or "startI3.t" for the
             *   introductory starter file 
             */
            GetModuleFileName(0, fbuf, sizeof(fbuf));
            strcpy(os_get_root_name(fbuf),
                   advanced ? "startA3.t" : "startI3.t");
        }

        /* copy the file */
        if (!CopyFile(fbuf, srcfile, FALSE))
        {
            sprintf(msg, "An error occurred creating the new game's "
                    "source file (%s).  Please make sure the folder "
                    "is correct and the disk isn't full.", srcfile);
            MessageBox(handle_, msg, "TADS Workbench",
                       MB_ICONEXCLAMATION | MB_OK);

            /* abort */
            return;
        }
    }

    /*
     *   Forget the old configuration file, since we're about to start
     *   modifying the configuration object for the new configuration.
     *   We've already saved the old file, so we don't need to know about
     *   it any more, and we want to be sure we don't try to save the new
     *   configuration to the old file.  For now, set an empty
     *   configuration; this will ensure that we don't save the
     *   configuration at any point (since it's empty), and that we load a
     *   new configuration (since the new config filename will be
     *   different when we load the game than the current config
     *   filename).  
     */
    local_config_filename_.set(0, 0);

    /* 
     *   Set up the new configuration, based on the existing
     *   configuration.  First, clear out all of the information that
     *   isn't suitable for a default configuration, so that we carry
     *   forward only the information from this configuration that we'd
     *   apply to the default config.  
     */
    clear_non_default_config();

    /* set up the initial build configuration for the new game */
    set_dbg_build_defaults(this, gamefile);
    local_config_->setval("build", "source_files", 0, srcfile);

    /*
     *   Build our source list: include the source file, plus the files it
     *   includes (adv.t and std.t).  First, clear out the source list.  
     */
    helper_->delete_internal_sources();

    /* add our starter source file */
    helper_->add_internal_line_source(os_get_root_name((textchar_t *)srcfile),
                                      srcfile, 0);

    /* add tads.h to the include files */
    GetModuleFileName(0, fbuf, sizeof(fbuf));
    strcpy(os_get_root_name(fbuf), "tads.h");
    oss_get_rel_path(rel_path, sizeof(rel_path), fbuf, proj_path);
    local_config_->setval("build", "include_files", 0, rel_path);

    /* add t3.h to the include files */
    strcpy(os_get_root_name(rel_path), "t3.h");
    local_config_->setval("build", "include_files", 1, rel_path);

    /* ask the helper to save the new source list into our config object */
    helper_->save_internal_source_config(local_config_);

    /* add the source directory to the directory list */
    os_get_path_name(fbuf, sizeof(fbuf), srcfile);
    local_config_->setval("srcdir", 0, 0, fbuf);

    /* 
     *   add the TADS executable directory to the directory list (for
     *   system header files) 
     */
    GetModuleFileName(0, msg, sizeof(msg));
    os_get_path_name(fbuf, sizeof(fbuf), msg);
    local_config_->setval("srcdir", 0, 1, fbuf);

    /* save the new configuration */
    local_config_->save_file(new_config);

    /*
     *   Set up the event loop flags so that we terminate the current
     *   event loop and load the new game 
     */
    reloading_new_ = TRUE;
    restarting_ = TRUE;
    reloading_ = TRUE;
    reloading_go_ = FALSE;
    reloading_exec_ = FALSE;

    /* tell the main loop to give up control so that we load the new game */
    go_ = TRUE;

    /* set the new game to load */
    if (CHtmlSys_mainwin::get_main_win() != 0)
        CHtmlSys_mainwin::get_main_win()->set_pending_new_game(gamefile);

    /*
     *   If we created the source file, post a Compile command, so that we
     *   compile the game as soon as we finish loading it.  (The game
     *   should be fully loaded by the time we get around to processing
     *   events again.)  Don't do this if we didn't create the source
     *   file, since we're not sure it's in any condition to compile at
     *   the moment.  
     */
    if (create_source)
        PostMessage(handle_, WM_COMMAND, ID_BUILD_COMPDBG, 0);

    /*
     *   Post an Open File command to open the first source file, which is
     *   the main game source file. 
     */
    PostMessage(handle_, WM_COMMAND, ID_FILE_OPENLINESRC_FIRST, 0);
}

/* ------------------------------------------------------------------------ */
/*
 *   Clear non-default configuration data 
 */
void CHtmlSys_dbgmain::vsn_clear_non_default_config()
{
    /* clear all project information */
    CHtmlDbgProjWin::clear_proj_config(local_config_);

    /* reset the project window */
    if (projwin_ != 0)
        projwin_->reset_proj_win();
}

/*
 *   Set version-specific build defaults 
 */
void vsn_set_dbg_build_defaults(CHtmlDebugConfig *config,
                                CHtmlSys_dbgmain *dbgmain,
                                const textchar_t *gamefile)
{
}

/* ------------------------------------------------------------------------ */
/*
 *   Run the debugger about box 
 */
void CHtmlSys_dbgmain::vsn_run_dbg_about()
{
    CTadsWinAboutTads3 *dlg;

    /* 
     *   create the window - it'll take care of the rest (including
     *   deleting itself when the user closes it) 
     */
    dlg = new CTadsWinAboutTads3(IDB_ABOUT_TDB3, w32t3_credits_txt);
    dlg->create_system_window(0, handle_, TRUE, "About TADS", 0,
                              new CTadsSyswin(dlg));
}

/* ------------------------------------------------------------------------ */
/*
 *   Program arguments dialog 
 */

class CTadsDlg_args: public CTadsDialog
{
public:
    CTadsDlg_args(CHtmlSys_dbgmain *dbgmain)
    {
        /* remember the debugger object */
        dbgmain_ = dbgmain;
    }
    
    /* initialize */
    void init();

    /* execute a command */
    int do_command(WPARAM id, HWND ctl);

protected:
    /* save our configuration settings */
    void save_config();
    
    /* debugger main window */
    CHtmlSys_dbgmain *dbgmain_;
};

/*
 *   initialize 
 */
void CTadsDlg_args::init()
{
    CHtmlDebugConfig *config;
    char buf[1024];

    /* get the local configuration object */
    config = dbgmain_->get_local_config();

    /* retrieve our command file setting and initialize the field */
    if (!config->getval("prog_args", "cmdfile", 0, buf, sizeof(buf)))
        SetWindowText(GetDlgItem(handle_, IDC_FLD_CMDFILE), buf);
    
    /* retrieve our log file setting and initialize the field */
    if (!config->getval("prog_args", "logfile", 0, buf, sizeof(buf)))
        SetWindowText(GetDlgItem(handle_, IDC_FLD_LOGFILE), buf);

    /* retrieve our arguments setting and initialize the field */
    if (!config->getval("prog_args", "args", 0, buf, sizeof(buf)))
        SetWindowText(GetDlgItem(handle_, IDC_FLD_PROGARGS), buf);

    /* inherit default */
    CTadsDialog::init();
}

/*
 *   save our configuration settings 
 */
void CTadsDlg_args::save_config()
{
    CHtmlDebugConfig *config;
    char buf[1024];

    /* get the local configuration object */
    config = dbgmain_->get_local_config();

    /* save the command file setting */
    GetWindowText(GetDlgItem(handle_, IDC_FLD_CMDFILE), buf, sizeof(buf));
    config->setval("prog_args", "cmdfile", 0, buf);

    /* save the log file setting */
    GetWindowText(GetDlgItem(handle_, IDC_FLD_LOGFILE), buf, sizeof(buf));
    config->setval("prog_args", "logfile", 0, buf);

    /* save the program arguments setting */
    GetWindowText(GetDlgItem(handle_, IDC_FLD_PROGARGS), buf, sizeof(buf));
    config->setval("prog_args", "args", 0, buf);
}

/*
 *   process a command 
 */
int CTadsDlg_args::do_command(WPARAM id, HWND ctl)
{
    OPENFILENAME info;
    char buf[1024];
    char fname[MAX_PATH];
    int result;
    HWND file_fld;
    int file_read;
    char *title;
    char path[MAX_PATH];

    /* see what we have */
    switch(id)
    {
    case IDC_BTN_SEL_CMDFILE:
        file_fld = GetDlgItem(handle_, IDC_FLD_CMDFILE);
        file_read = TRUE;
        title = "Select a command input file to read";
        goto ask_for_file;
        
    case IDC_BTN_SEL_LOGFILE:
        file_fld = GetDlgItem(handle_, IDC_FLD_LOGFILE);
        file_read = FALSE;
        title = "Select a log file for capturing output";

    ask_for_file:
        /* set up the open-file structure */
        info.lStructSize = sizeof(info);
        info.hwndOwner = handle_;
        info.hInstance = CTadsApp::get_app()->get_instance();
        info.lpstrFilter = "Text Files\0*.txt\0All Files\0*.*\0";
        info.lpstrCustomFilter = 0;
        info.nFilterIndex = 0;
        info.lpstrFile = buf;
        info.nMaxFile = sizeof(buf);
        info.lpstrFileTitle = 0;
        info.nMaxFileTitle = 0;
        info.lpstrTitle = title;
        info.Flags = OFN_NOCHANGEDIR;
        info.nFileOffset = 0;
        info.nFileExtension = 0;
        info.lpstrDefExt = 0;
        info.lCustData = 0;
        info.lpfnHook = 0;
        info.lpTemplateName = 0;
        CTadsDialog::set_filedlg_center_hook(&info);

        /* get the original filename from the field */
        GetWindowText(file_fld, buf, sizeof(buf));

        /* 
         *   get the path to the game file, so that we can make our
         *   filename relative to this path 
         */
        os_get_path_name(path, sizeof(path),
                         dbgmain_->get_gamefile_name());

        /* 
         *   if the filename stored here is relative, build the full path
         *   relative to the directory containing the current image file 
         */
        if (!os_is_file_absolute(buf))
        {
            /* build the full path to the file */
            os_build_full_path(fname, sizeof(fname), path, buf);

            /* copy the result back to the dialog's filename buffer */
            strcpy(buf, fname);
        }

        /* start off in the directory containing the file */
        info.lpstrInitialDir = 0;

        /* ask for the file using the appropriate dialog */
        if (file_read)
            result = GetOpenFileName(&info);
        else
            result = GetSaveFileName(&info);

        /* if it succeeded, store the result */
        if (result != 0)
        {
            
            /* 
             *   translate the filename back to relative form, relative to
             *   the directory containing the game file 
             */
            oss_get_rel_path(fname, sizeof(fname), buf, path);
            
            /* copy the new filename back into the text field */
            SetWindowText(file_fld, fname);

            /* update the open-file directory */
            CTadsApp::get_app()->set_openfile_dir(buf);
        }

        /* handled */
        return TRUE;

    case IDOK:
        /* save the settings */
        save_config();
        
        /* proceed to default processing */
        break;

    default:
        /* use default processing */
        break;
    }

    /* inherit default processing */
    return CTadsDialog::do_command(id, ctl);
}

/*
 *   run the program argument dialog 
 */
void CHtmlSys_dbgmain::vsn_run_dbg_args_dlg()
{
    CTadsDlg_args *dlg;

    /* create the dialog */
    dlg = new CTadsDlg_args(this);

    /* run the dialog */
    dlg->run_modal(DLG_PROGARGS, get_handle(),
                   CTadsApp::get_app()->get_instance());

    /* delete the dialog */
    delete dlg;
}

